function d = YAFIR(varargin)
    pIn = vararginToStruct(varargin);
    d = struct('legList', {{}}, 'nextPlotIx', 700); % data structure will become object
    
    % ********************************************
    % process input arguments that become part of the object
    % ********************************************

    % sampling rate at input
    [d, pIn] = moveParam(d, pIn, 'inRate_Hz');

    % number of taps
    [d, pIn] = moveParam(d, pIn, 'mStages');
    
    % number of frequency points. Increase to improve accuracy.
    % Frequencies are quantized to +/- rate / (2 * nSamples) 
    [d, pIn] = moveParam(d, pIn, 'nSamples', 1024);
    
    % default polyphase interpolation: none 
    [d, pIn] = moveParam(d, pIn, 'fInterp', 1);
    
    % default polyphase decimation: none
    [d, pIn] = moveParam(d, pIn, 'fDecim', 1);

    % max. number of iterations
    [d, pIn] = moveParam(d, pIn, 'nIter', 100);

    % for pure LMS solution, set nIter to 1 and change the weights below as needed
    [d, pIn] = moveParam(d, pIn, 'inbandWeight_initValue', 1);
    [d, pIn] = moveParam(d, pIn, 'outOfBandWeight_initValue', 1);

    % abort when the iteration weights grow too large. 
    % This happens when targets are impossible.
    % The result may still be meaningful, though.
    [d, pIn] = moveParam(d, pIn, 'abortWeight', 1e12);
    
    % keep iterating, if the targets are reached. 
    % Once the "equi"-ripple iteration has brought all peaks to an acceptable level, 
    % the LMS solver will use the remaining "degrees of freedom" for a LMS optimization.
    % The solver improves "where it is easy / cheap". This results in sloped
    % stopbands and "drooping" EVM in passbands.
    % Often, LMS is the best choice => set converged iterations to 0.
    [d, pIn] = moveParam(d, pIn, 'convergedIterations', 10);
    
    % for a complex-valued filter, use "true".
    % With a real-valued design, user input is only evaluated for positive frequencies!
    [d, pIn] = moveParam(d, pIn, 'complexValued', false);

    % false: the frequency response to be equalized applies only in the passband
    % true: stopband requirements are adjusted by the equalization target
    % for example: if EQ target provides 10 dB attenuation at some frequency, the stopband requirement will be relaxed by 10 dB
    [d, pIn] = moveParam(d, pIn, 'stopbandRequirementIncludesEq', false);

    % by default, the basis waveforms given to the optimizer are
    % within a delay range of +/- half the FIR length.
    % For nonlinear phase types (equalization / nominal frequency 
    % response), this may be suboptimal.
    % Meaningful values shouldn't exceed +/- half the number of
    % coefficients in the impulse response.
    [d, pIn] = moveParam(d, pIn, 'delayOffset', false);

    % describes a frequency response that is to be equalized in the passband
    % according to the EVM or ripple requirement
    [d, pIn] = moveParam(d, pIn, 'equalizeFreq_Hz', []);
    [d, pIn] = moveParam(d, pIn, 'equalizeGain_dB', []);
    [d, pIn] = moveParam(d, pIn, 'equalizeComplexGain', []);

    % describes a frequency response that is to be implemented in the passband
    % according to the EVM or ripple requirement
    [d, pIn] = moveParam(d, pIn, 'nominalFreq_Hz', []);
    [d, pIn] = moveParam(d, pIn, 'nominalGain_dB', []);
    [d, pIn] = moveParam(d, pIn, 'nominalComplexGain', []);
    
    % ********************************************
    % process input arguments that do not become part of the object
    % ********************************************    
    nonObjectArgs = struct();
    [nonObjectArgs, pIn] = moveParam(nonObjectArgs, pIn, 'req_passbandF_Hz', []);
    [nonObjectArgs, pIn] = moveParam(nonObjectArgs, pIn, 'req_passbandRipple_dB', []);
    [nonObjectArgs, pIn] = moveParam(nonObjectArgs, pIn, 'req_passbandEVM_dB', []);
    [nonObjectArgs, pIn] = moveParam(nonObjectArgs, pIn, 'req_stopbandF_Hz', []);
    [nonObjectArgs, pIn] = moveParam(nonObjectArgs, pIn, 'req_stopbandMaxGain_dB', []);
    
    % ********************************************
    % Input argument handling ends
    % all data has been copied. Any arguments that 
    % remain in pIn are faulty.
    % ********************************************
    assertEmptyArgumentsStruct(pIn);
    
    % ********************************************
    % object field fb:
    % construct frequency base for each FFT bin
    % ********************************************
    fb = 0:(d.nSamples - 1);
    fb = fb + floor(d.nSamples / 2);
    fb = mod(fb, d.nSamples);
    fb = fb - floor(d.nSamples / 2);
    fb = fb / d.nSamples; % now [0..0.5[, [-0.5..0[
    fb = fb * d.inRate_Hz * d.fInterp;
    d.fb = fb;
    
    % ********************************************
    % object field fbUser:
    % construct frequency base for the evaluation of 
    % given frequency responses
    % in real-valued mode, negative frequencies are 
    % converted to positive
    % ********************************************
    if d.complexValued
        d.fbUser = fb;
    else
        d.fbUser = abs(fb);
    end
    
    % ********************************************
    % object fields:
    % target settings. Those will be modified when
    % passbands / stopbands are placed
    % ********************************************
    % initial value of NaN indicates: all entries are unset
    d.errorSpecBinVal_inband_dB = zeros(size(d.fb)) + NaN;
    d.errorSpecBinVal_outOfBand_dB = zeros(size(d.fb)) + NaN;
    d.errorSpecBinVal_outOfBandOrig_dB = [];
    
    % ********************************************
    % Create object
    % at this point, all possible object fields must
    % have been assigned a value.
    % ********************************************
    d = class(d, 'YAFIR');
    
    % ********************************************
    % process req_passband requirement
    % needs to be done at setup, because the field is 
    % used for 'gating' with the tightenExisting / 
    % relaxExisting options
    % ********************************************
    if ~isempty(nonObjectArgs.req_passbandF_Hz)
        par = struct('onOverlap', 'error');
        if ~isempty(nonObjectArgs.req_passbandRipple_dB)
            par.ripple_dB = nonObjectArgs.req_passbandRipple_dB;
        end
        if ~isempty(nonObjectArgs.req_passbandEVM_dB)
            par.EVM_dB = nonObjectArgs.req_passbandEVM_dB;
        end
        par.f_Hz = nonObjectArgs.req_passbandF_Hz;

        d = placePassband(d, par);
    end % if req_passbandF_Hz

    % ********************************************
    % process req_stopband requirement
    % needs to be done now, because the field is 
    % used for 'gating' with the tightenExisting / 
    % relaxExisting options
    % ********************************************
    if ~isempty(nonObjectArgs.req_stopbandF_Hz)
        f_Hz = nonObjectArgs.req_stopbandF_Hz;
        g_dB = nonObjectArgs.req_stopbandMaxGain_dB;

        % extend to infinity
        if isscalar(f_Hz)
            f_Hz = [f_Hz 9e19]; 
            g_dB = [g_dB g_dB(end)];
        end
        
        d = placeBand...
            (d, ...
             'f_Hz', f_Hz, 'g_dB', g_dB, ...
             'type', 'stopband', ...
             'onOverlap', 'tighten');
    end
end

% ********************************************
% arguments handler, helper functions
% ********************************************
function [d, pIn] = moveParam(d, pIn, fieldname, defaultVal);
    if isfield(pIn, fieldname)
        d.(fieldname) = pIn.(fieldname);
        pIn = rmfield(pIn, fieldname);
    else
        if nargin < 4
            error(sprintf('argument "%s" is mandatory', fieldname));
        end
        d.(fieldname) = defaultVal;
    end
end

function [d, pIn] = assertEmptyArgumentsStruct(pIn);
    n = fieldnames(pIn);
    if numel(n) > 0
        for ix = 1:numel(n)
            disp(sprintf('unknown argument: %s', n{ix}));
        end
        error('aborting');
    end
end
